.586
.model flat,stdcall
option casemap:none
.code

CPP_COMPATIBLE = 1 ; main switch

; !! rebuilt for C/C++ !!
; original version had pRWL in ECX

;-----------------[ Notes on performance-customization ]------------------------\
;
;	LockByte_multiCPU   
;		Can be made to use one "rep nop" (P4's pause) instruction,
;		instead of the current 4 NOPs. Do this only if you are targeting
;		only one type of CPU, and you know how many cycles the "rep nop"
;		will take. LockByte_multiCPU should do a spinlock for 30-52 cycles,
;		(not accounting for the time lock cmpxchg takes).
;
;	@UnlockByte
;		There are two ways of implementing it: 
;		1) via xchg - will lock the bus. This might be mostly useful if the 
;			same RWLock object usually gets more than 
;			(very roughly estimated) 100,000 hits per second. 			
;		2) via mov (currently selected approach) - it will rely on 
;			the multiprocessors' cache-coherency mechanism. 
;			Saves 18 cycles on the current CPU, and a Bus-lock
;			for the other cpus.  
;-------------------------------------------------------------------------------/

;-------[ fixes]---------------------------------\
; 06.Mar.2007: LOCK was not added before cmpxchg, <- thanks to tbaust for noting
; 06.Mar.2007: StartWrite lockdown fixed <- thanks to dogby and pocjoc for noting
;------------------------------------------------/



;----------[ exports, defines ]----------\

RWLOCK struct
	Lock1		db ? ; [0..1]
	numWriters	db ? ; [0..1]
	numReaders	dw ? ; [0..65535]
RWLOCK ends ; 4 bytes

;----------------------------------------/


if CPP_COMPATIBLE
	pRWL textequ <pRWLock>  ; this is the parameter, passed to procs
	
	@GetECX macro
		mov ecx,[esp+4]
	endm	
	
	return macro
		ret 4
	endm
	
	; these procs require _this in the first param
	RWLock_Init proto pThis:DWORD
	RWLock_Free proto pThis:DWORD
	
	RWLock_StartRead proto 		pThis:DWORD
	RWLock_StartWrite proto 	pThis:DWORD
	
	RWLock_TryStartRead proto 	pThis:DWORD
	RWLock_TryStartWrite proto 	pThis:DWORD
	
	RWLock_EndRead proto 		pThis:DWORD
	RWLock_EndWrite proto 		pThis:DWORD
	RWLock_EndAccess proto 		pThis:DWORD
	
	RWLock_UpgradeToWriter proto 	pThis:DWORD
	RWLock_DowngradeToReader proto 	pThis:DWORD
	
	
else
	pRWL textequ <>
	
	@GetECX macro
	endm
	
	return macro
		ret
	endm
	
	; these procs require _this in ECX
	RWLock_Init proto
	RWLock_Free proto
	
	RWLock_StartRead proto
	RWLock_StartWrite proto
	
	RWLock_TryStartRead proto
	RWLock_TryStartWrite proto
	
	RWLock_EndRead proto
	RWLock_EndWrite proto
	RWLock_EndAccess proto
	
	RWLock_UpgradeToWriter proto
	RWLock_DowngradeToReader proto
	
endif

RWLock_MultiStartWrite PROTO C NumParam:DWORD,Params:VARARG
RWLock_MultiEndWrite   PROTO C NumParam:DWORD,Params:VARARG

OSVERSIONINFOA STRUCT
  dwOSVersionInfoSize   DWORD      ?
  dwMajorVersion        DWORD      ?
  dwMinorVersion        DWORD      ?
  dwBuildNumber         DWORD      ?
  dwPlatformId          DWORD      ?
  szCSDVersion          BYTE 128 dup (?)
OSVERSIONINFOA ENDS

SYSTEM_INFO STRUCT
  wProcessorArchitecture        WORD       ?
  wReserved                     WORD       ?
  dwPageSize                    DWORD      ?
  lpMinimumApplicationAddress   DWORD      ?
  lpMaximumApplicationAddress   DWORD      ?
  dwActiveProcessorMask         DWORD      ?
  dwNumberOfProcessors         DWORD      ?
  dwProcessorType               DWORD      ?
  dwAllocationGranularity       DWORD      ?
  wProcessorLevel               WORD       ?
  wProcessorRevision            WORD       ?
SYSTEM_INFO ENDS

Sleep proto :DWORD
SwitchToThread proto
GetSystemInfo proto :DWORD
GetVersionExA proto :DWORD


.data
	align 16
	LockByteProc dd 0
	SwitchThreadProc dd 0
.code

@TRYLOCK macro
	mov al,0
	mov ah,1
	lock cmpxchg [ecx],ah  ; compares [ecx] with al, and if equal, sets [ecx]=ah
	;; if [ecx] was 0, zero-flag is set *** -> good outcome
	;; if [ecx] was 1, not zero flag
	;; al= previous value of [ecx]
endm


LockByte_oneCPU:
	call SwitchThreadProc
	@TRYLOCK
	jnz LockByte_oneCPU
	ret
	
LockByte_multiCPU:
	mov dword ptr[esp-4],20 ; set spin-count
	@@: ; next spin 
	dec dword ptr[esp-4]
	nop
	nop
	nop
	nop
	jz @F ; spin-count has ended, we must switch-thread
	@TRYLOCK
	jnz @B
	ret
	@@: ; do switch-thread, since spin-count ended
	call SwitchThreadProc
	@TRYLOCK
	jnz LockByte_multiCPU ; try again with spin-count if failed to lock now
	ret
	
SwitchThread_win98:
	pushad
	invoke Sleep,0
	popad
	ret
SwitchThread_win2K:
	pushad
	invoke SwitchToThread
	popad
	ret
	
@LockByte macro
	@TRYLOCK
	.if !ZERO? ; if not locked
		call LockByteProc
	.endif
endm

@UnlockByte macro 
	;------[ version 1]----[
	;xor al,al
	;lock xchg [ecx],al
	;----------------------/
	
	;----[ version 2 ]--[
	mov byte ptr[ecx],0
	;-------------------/
endm

@SwitchThread macro
	call SwitchThreadProc
endm
	





option PROLOGUE:NONE
option EPILOGUE:NONE


align 16
RWLock_StartRead proc PUBLIC pRWL
	@GetECX
	again:
	@LockByte
	.if ![ecx].RWLOCK.numWriters
		inc [ecx].RWLOCK.numReaders
		@UnlockByte
		return
	.endif
	
	@UnlockByte
	@SwitchThread
	jmp again
RWLock_StartRead endp

align 16
RWLock_EndRead proc PUBLIC pRWL
	@GetECX
	@LockByte
	dec [ecx].RWLOCK.numReaders
	@UnlockByte
	return
RWLock_EndRead endp

align 16
RWLock_StartWrite proc PUBLIC pRWL
	@GetECX
	again:
	@LockByte
	.if [ecx].RWLOCK.numWriters==0
		mov [ecx].RWLOCK.numWriters,1 ; reserve :)
		checkReaders:
		.if [ecx].RWLOCK.numReaders==0
			@UnlockByte
			return
		.endif
		@UnlockByte
		@SwitchThread
		;----[ @LockByte ]---------[
		@TRYLOCK
		jz checkReaders; if locked 
		call LockByteProc
		;--------------------------/
		jmp checkReaders
	.endif
	@UnlockByte
	@SwitchThread
	jmp again
RWLock_StartWrite endp

	
align 16
RWLock_EndWrite proc PUBLIC pRWL
	@GetECX
	@LockByte
	mov [ecx].RWLOCK.numWriters,0
	@UnlockByte
	return
RWLock_EndWrite endp

align 16
RWLock_EndAccess proc PUBLIC pRWL
	@GetECX
	@LockByte
	.if [ecx].RWLOCK.numReaders
		dec [ecx].RWLOCK.numReaders
	.else
		mov [ecx].RWLOCK.numWriters,0
	.endif
	@UnlockByte
	return
RWLock_EndAccess endp

align 16
RWLock_TryStartRead proc PUBLIC pRWL
	@GetECX
	@LockByte
	.if ![ecx].RWLOCK.numWriters
		inc [ecx].RWLOCK.numReaders
		@UnlockByte
		xor eax,eax
		inc eax
		return
	.endif
	@UnlockByte
	xor eax,eax
	return
RWLock_TryStartRead endp	


align 16
RWLock_TryStartWrite proc PUBLIC pRWL
	@GetECX
	again:
	@LockByte
	.if ![ecx].RWLOCK.numReaders && ![ecx].RWLOCK.numWriters
		mov [ecx].RWLOCK.numWriters,1
		@UnlockByte
		xor eax,eax
		inc eax
		return
	.endif
	@UnlockByte
	xor eax,eax
	return
RWLock_TryStartWrite endp

RWLock_UpgradeToWriter proc PUBLIC pRWL
	@GetECX
	again:
	@LockByte
	.if [ecx].RWLOCK.numReaders==1 && ![ecx].RWLOCK.numWriters
		mov [ecx].RWLOCK.numReaders,0
		mov [ecx].RWLOCK.numWriters,1
		@UnlockByte
		return
	.endif
	@UnlockByte
	@SwitchThread
	jmp again
RWLock_UpgradeToWriter endp

RWLock_DowngradeToReader proc PUBLIC pRWL
	@GetECX
	@LockByte
	mov [ecx].RWLOCK.numWriters,0
	inc [ecx].RWLOCK.numReaders
	@UnlockByte
	return
RWLock_DowngradeToReader endp



align 16
RWLock_Free proc PUBLIC pRWL
	@GetECX
	again:
	@LockByte
	.if [ecx].RWLOCK.numReaders==0 && [ecx].RWLOCK.numWriters==0
		return ; !! leaves the proc locked on purpose, to show the coder that he is doing something wrong
	.endif
	@UnlockByte
	jmp again		
RWLock_Free endp


align 16
RWLock_Init proc PUBLIC pRWL ; ecx = _this
	@GetECX
	;---[ clear the struct ]---\
	mov dword ptr[ecx],0
	;mov [ecx].RWLOCK.Lock1,0
	;mov [ecx].RWLOCK.numWriters,0
	;mov [ecx].RWLOCK.numReaders,0
	;--------------------------/
	.if LockByteProc
		return
	.endif
	;----[ set locking proc ]-------------\
	pushad
	sub esp,SYSTEM_INFO
	mov eax,esp
	invoke GetSystemInfo,eax
	mov LockByteProc,LockByte_multiCPU
	.if [esp].SYSTEM_INFO.dwNumberOfProcessors==1
		mov LockByteProc,LockByte_oneCPU
	.endif
	add esp,SYSTEM_INFO-OSVERSIONINFOA
	mov [esp].OSVERSIONINFOA.dwOSVersionInfoSize,OSVERSIONINFOA
	mov eax,esp
	invoke GetVersionExA,eax
	mov SwitchThreadProc,SwitchThread_win98
	.if [esp].OSVERSIONINFOA.dwPlatformId!=1
		mov SwitchThreadProc,SwitchThread_win2K
	.endif
	
	add esp,OSVERSIONINFOA
	popad
	;-------------------------------------/
	
	return
RWLock_Init endp

option PROLOGUE:PrologueDef
option EPILOGUE:EpilogueDef

RWLock_MultiEndWrite proc C PUBLIC uses edx esi NumParams,Params:VARARG
	mov edx,NumParams
	test edx,edx
	jz _ret
	
	;------[ select source ]-----\
	lea esi,Params
	.if SIGN? ; if edx<0
		neg edx
		neg NumParams
		mov esi,Params
	.endif
	;----------------------------/
	
	align 16
	next:
	dec edx
	mov ecx,[esi+edx*4]
	.if ecx
		@LockByte
		mov [ecx].RWLOCK.numWriters,0
		@UnlockByte
	.endif
	test edx,edx
	jnz next
	
_ret:	ret
RWLock_MultiEndWrite endp


RWLock_MultiStartWrite proc C PUBLIC uses ebx edx esi edi NumParams,Params:VARARG
	local credits[300]:BYTE
	
	mov edx,NumParams
	test edx,edx
	jz _ret
	
	;------[ select source ]-----\
	lea esi,Params
	.if SIGN? ; if edx<0
		neg edx
		neg NumParams
		mov esi,Params
	.endif
	;----------------------------/
	
	
	;----[ fill credits[] ]------\
	; give 2 credits for existing
	; give 0 for non-existing
	xor edi,edi ; NumTotalCredits
	xor ebx,ebx
	xor eax,eax
	@@:
	dec edx
	mov al,0
	.if [esi+edx*4]!=ebx
		mov al,2
	.endif
	mov credits[edx],al
	add edi,eax
	test edx,edx
	jnz @B
	;----------------------------/
	
    TryAgain:
	xor edx,edx
	.while edx<NumParams
		mov bl,credits[edx]
		mov ecx,[esi+edx*4]		
		.if bl && ecx ; if has credits and has valid pointer
			@LockByte
			cmp bl,1
			je _credits_1
			
			.if [ecx].RWLOCK.numWriters==0
				dec bl
				dec edi
				mov [ecx].RWLOCK.numWriters,1 ; reserve :)
				_credits_1:
				.if [ecx].RWLOCK.numReaders==0
					dec bl
					dec edi
				.endif
			.endif
			@UnlockByte
			mov credits[edx],bl
		.endif
		
		inc edx
	.endw
	.if !edi
		ret
	.endif
	@SwitchThread
	jmp TryAgain
	
	
_ret:	ret
RWLock_MultiStartWrite endp




end